@@ -17,7 +17,7 @@ from django.template.loader import render_to_string, get_template
 from django.utils.encoding import force_text, python_2_unicode_compatible
 from django.utils.html import conditional_escape, format_html
 from django.utils.safestring import mark_safe
-from django.db.models import QuerySet
+from django.db.models import QuerySet, warnings
 from six import string_types
 from tri.declarative import declarative, creation_ordered, with_meta, setdefaults, evaluate_recursive, evaluate, getattr_path, collect_namespaces, extract_subkeys, sort_after, LAST, setdefaults_path, dispatch, EMPTY
 from tri.form import Field, Form, member_from_model, expand_member, create_members_from_model
@@ -25,10 +25,18 @@ from tri.named_struct import NamedStructField, NamedStruct
 from tri.struct import Struct, Frozen, merged
 from tri.query import Query, Variable, QueryException, Q_OP_BY_OP
 
-__version__ = '2.0.0'
+from tri.table.db_compat import setup_db_compat
+
+__version__ = '2.5.0'
 
 LAST = LAST
 
+_column_factory_by_field_type = OrderedDict()
+
+
+def register_column_factory(field_class, factory):
+    _column_factory_by_field_type[field_class] = factory
+
 
 def prepare_headers(request, bound_columns):
     """
@@ -184,7 +192,7 @@ class Column(Frozen, ColumnBase):
         super(Column, self).__init__(**kwargs)
 
     @staticmethod
-    def text(**kwargs):
+    def text(**kwargs):  # pragma: no cover
         return Column(**kwargs)
 
     @staticmethod
@@ -322,10 +330,18 @@ class Column(Frozen, ColumnBase):
 
     @staticmethod
     def float(**kwargs):
+        setdefaults(kwargs, dict(
+            query__class=Variable.float,
+            bulk__class=Field.float,
+        ))
         return Column.number(**kwargs)
 
     @staticmethod
     def integer(**kwargs):
+        setdefaults(kwargs, dict(
+            query__class=Variable.integer,
+            bulk__class=Field.integer,
+        ))
         return Column.number(**kwargs)
 
     @staticmethod
@@ -371,6 +387,7 @@ class Column(Frozen, ColumnBase):
     def date(**kwargs):
         setdefaults(kwargs, dict(
             query__gui__class=Field.date,
+            # TODO: query__class=Variable.date,
             query__op_to_q_op=lambda op: {'=': 'exact', ':': 'contains'}.get(op) or Q_OP_BY_OP[op],
             bulk__class=Field.date,
         ))
@@ -379,9 +396,10 @@ class Column(Frozen, ColumnBase):
     @staticmethod
     def datetime(**kwargs):
         setdefaults(kwargs, dict(
-            query__gui__class=Field.date,
+            query__gui__class=Field.datetime,
+            # TODO: query__class=Variable.datetime,
             query__op_to_q_op=lambda op: {'=': 'exact', ':': 'contains'}.get(op) or Q_OP_BY_OP[op],
-            bulk__class=Field.date,
+            bulk__class=Field.datetime,
         ))
         return Column(**kwargs)
 
@@ -390,6 +408,7 @@ class Column(Frozen, ColumnBase):
         setdefaults(kwargs, dict(
             bulk__class=Field.email,
             # TODO: query__class=Variable.email,
+            query__gui__class=Field.email,
         ))
         return Column(**kwargs)
 
@@ -397,7 +416,8 @@ class Column(Frozen, ColumnBase):
     def from_model(model, field_name=None, model_field=None, **kwargs):
         return member_from_model(
             model=model,
-            factory_lookup=_column_factory_by_django_field_type,
+            factory_lookup=_column_factory_by_field_type,
+            factory_lookup_register_function=register_column_factory,
             field_name=field_name,
             model_field=model_field,
             defaults_factory=lambda model_field: {},
@@ -407,7 +427,7 @@ class Column(Frozen, ColumnBase):
     def expand_member(model, field_name=None, model_field=None, **kwargs):
         return expand_member(
             model=model,
-            factory_lookup=_column_factory_by_django_field_type,
+            factory_lookup=_column_factory_by_field_type,
             field_name=field_name,
             model_field=model_field,
             **kwargs)
@@ -496,6 +516,24 @@ class BoundCell(object):
 
         self.value = evaluate(bound_column.cell.value, table=bound_row.table, column=bound_column.column, row=bound_row.row)
 
+    @property
+    def attrs(self):
+        return evaluate_recursive(self.bound_column.cell.attrs, table=self.table, column=self.bound_column, row=self.row, value=self.value)
+
+    @property
+    def url(self):
+        url = self.bound_column.cell.url
+        if callable(url):
+            url = url(table=self.table, column=self.bound_column, row=self.row, value=self.value)
+        return url
+
+    @property
+    def url_title(self):
+        url_title = self.bound_column.cell.url_title
+        if callable(url_title):
+            url_title = url_title(table=self.table, column=self.bound_column, row=self.row, value=self.value)
+        return url_title
+
     def render(self):
         cell__template = self.bound_column.cell.template
         if cell__template:
@@ -504,24 +542,17 @@ class BoundCell(object):
             return format_html('<td{}>{}</td>', self.render_attrs(), self.render_cell_contents())
 
     def render_attrs(self):
-        attrs = evaluate_recursive(self.bound_column.cell.attrs, table=self.table, column=self.bound_column, row=self.row, value=self.value)
-        return render_attrs(attrs)
+        return render_attrs(self.attrs)
 
     def render_cell_contents(self):
         cell_contents = self.render_formatted()
 
-        cell__url = self.bound_column.cell.url
-        if callable(cell__url):
-            cell__url = cell__url(table=self.table, column=self.bound_column, row=self.row, value=self.value)
-
-        if cell__url:
-            cell__url_title = self.bound_column.cell.url_title
-            if callable(cell__url_title):
-                cell__url_title = cell__url_title(table=self.table, column=self.bound_column, row=self.row, value=self.value)
-
-            cell_contents = format_html('<a href="{}"{}>{}</a>',
-                                        mark_safe(cell__url),
-                                        mark_safe(' title=%s' % cell__url_title) if cell__url_title else '',
+        url = self.url
+        if url:
+            url_title = self.url_title
+            cell_contents = format_html('<a{}{}>{}</a>',
+                                        format_html(' href="{}"', url),
+                                        format_html(' title="{}"', url_title) if url_title else '',
                                         cell_contents)
         return mark_safe(cell_contents)
 
@@ -564,6 +595,10 @@ class Table(object):
         header__template = 'tri_table/table_header_rows.html'
         links__template = 'tri_table/links.html'
 
+        endpoint_dispatch_prefix = None
+        endpoint__query = lambda table, key, value: table.query.endpoint_dispatch(key=key, value=value) if table.query is not None else None
+        endpoint__bulk = lambda table, key, value: table.bulk_form.endpoint_dispatch(key=key, value=value) if table.bulk is not None else None
+
         model = None
 
     def __init__(self, data=None, request=None, columns=None, columns_dict=None, **kwargs):
@@ -584,7 +619,8 @@ class Table(object):
 
         self._has_prepared = False
 
-        if data is None:
+        if data is None:  # pragma: no cover
+            warnings.warn('deriving model from data queryset is deprecated, use Table.from_model', DeprecationWarning)
             assert 'model' in kwargs and kwargs['model'] is not None
             data = kwargs['model'].objects.all()
 
@@ -620,6 +656,7 @@ class Table(object):
 
         self.query_kwargs = extract_subkeys(kwargs, 'query')
         self.bulk_kwargs = extract_subkeys(kwargs, 'bulk')
+        self.endpoint = extract_subkeys(kwargs, 'endpoint')
 
     def _prepare_auto_rowspan(self):
         auto_rowspan_columns = [column for column in self.shown_bound_columns if column.auto_rowspan]
@@ -774,6 +811,7 @@ class Table(object):
             self.query = Query(
                 request=request,
                 variables=variables,
+                endpoint_dispatch_prefix='__'.join(part for part in [self.Meta.endpoint_dispatch_prefix, 'form'] if part is not None),
                 **self.query_kwargs
             )
             self.query_form = self.query.form() if self.query.variables else None
@@ -809,7 +847,7 @@ class Table(object):
             self.bulk_form = Form(
                 data=request.POST,
                 fields=bulk_fields,
-                endpoint_dispatch_prefix='bulk',
+                endpoint_dispatch_prefix='__'.join(part for part in [self.Meta.endpoint_dispatch_prefix, 'bulk'] if part is not None),
                 **self.bulk_kwargs) if bulk_fields else None
 
         self._prepare_auto_rowspan()
@@ -817,6 +855,10 @@ class Table(object):
         return headers, self.header_levels
 
     def bound_rows(self):
+        return self
+
+    def __iter__(self):
+        self.prepare(self.request)
         for i, row in enumerate(self.data):
             yield BoundRow(table=self, row=row, row_index=i)
 
@@ -830,11 +872,11 @@ class Table(object):
     @staticmethod
     def columns_from_model(**kwargs):
         kwargs = collect_namespaces(kwargs)
-        kwargs['db_field'] = kwargs.pop('column', {})
+        kwargs['db_field'] = collect_namespaces(kwargs.pop('column', {}))
         return create_members_from_model(default_factory=Column.from_model, **kwargs)
 
     @staticmethod
-    def from_model(data, model, instance=None, include=None, exclude=None, extra_fields=None, post_validation=None, **kwargs):
+    def from_model(data=None, model=None, instance=None, include=None, exclude=None, extra_fields=None, post_validation=None, **kwargs):
         """
         Create an entire form based on the fields of a model. To override a field parameter send keyword arguments in the form
         of "the_name_of_the_field__param". For example:
@@ -851,14 +893,19 @@ class Table(object):
 
         """
         kwargs = collect_namespaces(kwargs)
+        assert model or data, "model or data must be specified"
+        if model is None and isinstance(data, QuerySet):
+            model = data.model
         columns = Table.columns_from_model(model=model, include=include, exclude=exclude, extra=extra_fields, column=kwargs.pop('column', {}))
         return Table(data=data, model=model, instance=instance, columns=columns, post_validation=post_validation, **kwargs)
 
     def endpoint_dispatch(self, key, value):
-        if key.startswith('query__'):
-            return self.query.endpoint_dispatch(key=key[len('query__'):], value=value)
-        if key.startswith('bulk__'):
-            return self.bulk.endpoint_dispatch(key=key[len('bulk__'):], value=value)
+        parts = key.split('__', 1)
+        prefix = parts.pop(0)
+        remaining_key = parts[0] if parts else None
+        for endpoint, handler in self.endpoint.items():
+            if prefix == endpoint:
+                return handler(table=self, key=remaining_key, value=value)
 
 
 class Link(Struct):
@@ -866,8 +913,8 @@ class Link(Struct):
     Class that describes links to add underneath the table.
     """
     # noinspection PyShadowingBuiltins
-    def __init__(self, title, url, show=True, group=None, id=None):
-        super(Link, self).__init__(title=title, url=url, show=show, group=group, id=id)
+    def __init__(self, title, url, show=True, group=None, id=None, **kwargs):
+        super(Link, self).__init__(title=title, url=url, show=show, group=group, id=id, **kwargs)
 
     @staticmethod
     def icon(icon, title, **kwargs):
@@ -997,8 +1044,6 @@ def render_table(request,
 
     if table is None or isinstance(table, dict):
         table_kwargs = table if isinstance(table, dict) else kwargs.pop('table', {})
-        if 'model' not in table_kwargs:
-            table_kwargs['model'] = table_kwargs['data'].model
         table = Table.from_model(**table_kwargs)
 
     table.prepare(request)
@@ -1006,7 +1051,15 @@ def render_table(request,
 
     for key, value in request.GET.items():
         if key.startswith('__'):
-            data = table.endpoint_dispatch(key=key[2:], value=value)
+            remaining_key = key[2:]
+            expected_prefix = table.Meta.endpoint_dispatch_prefix
+            if expected_prefix is not None:
+                parts = remaining_key.split('__', 1)
+                prefix = parts.pop(0)
+                if prefix != expected_prefix:
+                    return
+                remaining_key = parts[0] if parts else None
+            data = table.endpoint_dispatch(key=remaining_key, value=value)
             if data:
                 return HttpResponse(json.dumps(data), content_type='application/json')
 
@@ -1056,6 +1109,4 @@ def render_table_to_response(*args, **kwargs):
         return response
     return HttpResponse(response)
 
-from .db_compat import setup, _column_factory_by_django_field_type  # noqa
-
-setup()
+setup_db_compat()
